// This is my list manipulation package, modified to be thread safe,
// for use in creating a Scratch-like environment in C.  It can be used
// for a coroutine scheduler or for handling a stack eg for parameters
// or return addresses etc.

// cc -Wall -o dequeue dequeue.c -pthread

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>
#include <unistd.h> // if 'sleep()' used.

pthread_mutex_t listlock;
pthread_mutex_t iolock;

// As long as separate threads access separate lists, this is thread safe.
// No promises what happens if two threads manipulate the same list...

/* Double-ended queue implemented as circular list pointing to tail.

     empty queue is a ptr to NULL.
     queue with one item points to itself.
     queue with two or more items: queue points to last item,
                                   tail of last item points to head.

   Accessing the first or last item in the list is cheap.

   Prepending to (pushing on) or Appending to a list is cheap.

   Removing the head of a list (popping off) is cheap.

   Removing the last item from the list is expensive.

   Butfirst would be cheap if implemented.

   Butlast would be expensive.

   With the exception of an expensive 'butlast', this data structure is
   ideal for implementing lists in LOGO
   Commands needed:
     first butfirst last butlast member item setitem fput lput join
     count arraytolist listtoarray reverse combine

   These options efficiently support both stacks (LIFOs)
   and queues (FIFOs) with the same data structure.

   For a summary of different kinds of queues,
   see http://www.cis.upenn.edu/~matuszek/cit594-2005/Lectures/23-stacks-queues-deques.ppt

   Using a circular list avoids the cost of backpointers.

   At some point in the future there's a coding trick that can
   be added whereby the 'next' pointer is the exclusive-or of
   itself and the backpointer which somehow lets you store
   the backpointers without any extra space.  To be honest I
   don't remember the details and will have to sit down some time
   and work it out from first principles.  It's very hacky and
   unless removing the last item from a list is a frequent
   occurance in an application, it's not worth the increase in
   complexity of the code when this implementation is simple
   and works.

   (See http://en.wikipedia.org/wiki/XOR_linked_list )

   OK, the downside is that a list variable is no longer just
   one pointer, it has to be two pointers containing (in this case)
   the address of both the tail and the head of the list.

   It's too messy, I don't think I'll implement it.

   TO DO: an apply procedure, and a map function.  Remember a
   void * parameter for passing in arbitrary context.

 */

#ifndef USERTYPE
/* If usertype is anything other than a scalar the code will need
   tweaking wherever the data field is referenced. */
#define USERTYPE int
#endif

typedef struct QQ {
  struct QQ *tail;
  USERTYPE atom; /* more generically, a void *, but int is OK for testing */
} QQ;

static QQ *new_QQ(USERTYPE atom) /* returns a cell.  Used INTERNALLY only */
{
  QQ *p;
  pthread_mutex_lock(&listlock); // may not need lock here
  p = malloc(sizeof(QQ));
  assert(p != NULL);
  p->atom = atom;
  p->tail = NULL;
  pthread_mutex_unlock(&listlock);
  return p;
}

void append_QQ(QQ **pp, USERTYPE atom)
{
  QQ *head, *new;
#define p (*pp)
  if (p == NULL) {
    head = new_QQ(atom);
    pthread_mutex_lock(&listlock);
    p = head;
    p->tail = head;
  } else {
    /* p points to the tail. */
    head = p->tail;
    new = new_QQ(atom);
    pthread_mutex_lock(&listlock);
    p->tail = new;
    p = p->tail;
    p->tail = head;
  }
  pthread_mutex_unlock(&listlock);
#undef p
}
#define appendq(q, i) append_QQ(&q, i)

void push_QQ(QQ **pp, USERTYPE atom)
{
  QQ *head, *new;
#define p (*pp)
  if (p == NULL) {
    p = new_QQ(atom);
    pthread_mutex_lock(&listlock);
    p->tail = p;
  } else {
    /* p points to the tail. */
    head = p->tail;
    new = new_QQ(atom);
    pthread_mutex_lock(&listlock);
    p->tail = new;
    p->tail->tail = head;
  }
  pthread_mutex_unlock(&listlock);
#undef p
}
#define pushq(q, i) push_QQ(&q, i)

void show_QQ(int pid, char *prefix, QQ *list, const char *name)
{
  fprintf(stdout, "[%d] %s%s:", pid, prefix, name);
  if (list == NULL) {
    fprintf(stdout, " NULL");
  } else {
    QQ *loop_breaker = list;
    for (;;) {
      list = list->tail;
      fprintf(stdout, " %d", list->atom);
      if (list == loop_breaker) break;
    }
  }
  fprintf(stdout, "\n");
}
#define showq2(pid, prefix, name) show_QQ(pid, prefix, name, #name)
#define showq(pid, name) show_QQ(pid, "", name, #name)

int firstq(QQ *q)
{
  assert(q != NULL); /* user bug */
  assert(q->tail != NULL); /* package bug */
  return q->tail->atom;
}

int lastq(QQ *q)
{
  assert(q != NULL);
  return q->atom;
}

void popfirst_QQ(QQ **qq)
{
  QQ *head;
#define q (*qq)
  pthread_mutex_lock(&listlock);
  assert(qq != NULL);
  assert(q != NULL);
  head = q->tail;
  if (q == head) { /* last atom */
    q = NULL;
  } else {
    q->tail = q->tail->tail;
  }
  free(head);
  pthread_mutex_unlock(&listlock);
#undef q
}
#define popfirstq(q) popfirst_QQ(&q)

void poplast_QQ(QQ **qq)
{
  QQ *tail = NULL;
#define q (*qq)
  pthread_mutex_lock(&listlock);
  assert(qq != NULL);
  assert(q != NULL);

  if (q->tail != q) {
    /* EXPENSIVE */
    tail = q->tail;
    for (;;) {
      if (tail->tail == q) break;
      tail = tail->tail;
    }
    tail->tail = q->tail;
  }
  free(q); q = tail;
  pthread_mutex_unlock(&listlock);
#undef q
}
#define poplastq(q) poplast_QQ(&q)

/* Unsafe join (aka 'cons') of two lists */
QQ *destructive_join_QQ(QQ **lleft, QQ **rright)
{
  QQ *tmp = NULL;
#define left (*lleft)
#define right (*rright)
  pthread_mutex_lock(&listlock);
  /* join two lists.  Use their space.  Set the lists to NULL. */
  tmp = right->tail;
  right->tail = left->tail;
  left->tail = tmp;
  tmp = right;
  left = NULL; right = NULL;
  pthread_mutex_unlock(&listlock);
  return tmp;
#undef left
#undef right
}
#define destructive_joinq(left, right) destructive_join_QQ(&left, &right)

/* duplicate (copy) a list */
QQ *dupq(QQ *list)
{
  QQ *new = NULL, *old = list;
  USERTYPE atom;
  if (list != NULL) {
    for (;;) {
      atom = firstq(list); // list->tail->atom
      appendq(new, atom); // is a 3-liner in-line better than appendq?
      list = list->tail;
      if (list == old) break;
    }
  }
  return new;
}

/* reverse a copy of a list */
QQ *reversedupq(QQ *list)
{
  QQ *new = NULL, *old = list;
  USERTYPE atom;
  if (list != NULL) {
    for (;;) {
      atom = firstq(list); // list->tail->atom
      pushq(new, atom); // is a 3-liner in-line better than pushq?
      list = list->tail;
      if (list == old) break;
    }
  }
  return new;
}

/* count the elements in a list */
int countq(QQ *list)
{
  QQ *start = list;
  int count = 0;

  if (list != NULL) {
    for (;;) {
      list = list->tail; count += 1;
      if (list == start) break;
    }
  }
  return count;
}

/* return an indexed member of a list (based at 1) */
int memberq(QQ *list, int member)
{
  QQ *start = list;
  int count = 0;

  assert(member > 0);
  assert(list != NULL);
  list = list->tail;
  for (count = 1; count < member; count++) {
    assert(list != start); // not enough members
    list = list->tail;
  }

  return list->atom;
}

/* non-destructive join.  Uses copies of the lists. */
QQ *joinq(QQ *left, QQ *right)
{
  QQ *temp1 = dupq(left), *temp2 = dupq(right);
  return destructive_joinq(temp1, temp2);
}

/* use this to avoid heap lossage */
void make_QQ(QQ **ddest, QQ **ssource)
{
#define dest (*ddest)
#define source (*ssource)
  assert(ddest != NULL);
  while (dest != NULL) (void)popfirstq(dest); // 1-liner, no need for delq()  
  dest = source;
  source = NULL; /* Ensures we don't have two pointers to the one area. */
#undef dest
#undef source
}
#define makeq(dest, source) make_QQ(&dest, &source)

#define P pthread_mutex_lock(&iolock)
#define V pthread_mutex_unlock(&iolock)
void *list_test(void* state)
{
  int *my_clone_id_ptr = (int *)state;
  QQ *list1 = NULL, *list2 = NULL, *joined = NULL, *copied = NULL, *reversed = NULL;
  int my_clone_id = *my_clone_id_ptr;

  P;showq(my_clone_id, list1);V;
  appendq(list1, 21);
  P;showq(my_clone_id, list1);V;
  appendq(list1, 42);
  P;showq(my_clone_id, list1);V;
  appendq(list1, 84);
  P;showq(my_clone_id, list1);V;

  P;
  fprintf(stdout, "[%d] first(list1) = %d\n", my_clone_id, firstq(list1));
  fprintf(stdout, "[%d] last(list1) = %d\n", my_clone_id, lastq(list1));
  V;
  popfirstq(list1); P; showq2(my_clone_id, "popfirstq ", list1); V;
  popfirstq(list1); P; showq2(my_clone_id, "popfirstq ", list1); V;
  popfirstq(list1); P; showq2(my_clone_id, "popfirstq ", list1); V;

  P;showq(my_clone_id, list2);V;
  pushq(list2, 17);
  P;showq(my_clone_id, list2);V;
  pushq(list2, 34);
  P;showq(my_clone_id, list2);V;
  pushq(list2, 51);
  P;showq(my_clone_id, list2);V;

  P;
  fprintf(stdout, "[%d] first(list2) = %d\n", my_clone_id, firstq(list2));
  fprintf(stdout, "[%d] last(list2) = %d\n", my_clone_id, lastq(list2));
  V;
  poplastq(list2); P; showq2(my_clone_id, "poplastq ", list2);V;
  poplastq(list2); P; showq2(my_clone_id, "poplastq ", list2);V;
  poplastq(list2); P; showq2(my_clone_id, "poplastq ", list2);V;

  // rebuild
  appendq(list1, 21);
  appendq(list1, 42);
  appendq(list1, 84);
  P;showq(my_clone_id, list1);V;
  pushq(list2, 17);
  pushq(list2, 34);
  pushq(list2, 51);
  P;showq(my_clone_id, list2);V;
  joined = destructive_joinq(list1, list2);
  P;showq(my_clone_id, joined);V;
  P;showq(my_clone_id, list1);showq(my_clone_id, list2);V;

  copied = dupq(joined);
  P;showq(my_clone_id, copied);V;

  while (copied != NULL) (void)popfirstq(copied); // 1-liner, no need for delq()
  P;showq(my_clone_id, copied);V;

  while (copied != NULL) (void)popfirstq(copied); 
  P;showq(my_clone_id, copied);V;

  P;showq(my_clone_id, joined);V;
  makeq(copied, joined);
  P;showq(my_clone_id, copied);V;
  P;showq(my_clone_id, joined);V;
  joined = joinq(copied, copied);
  P;showq(my_clone_id, copied);V;
  P;showq(my_clone_id, joined);V;

  reversed = reversedupq(joined);
  P;showq(my_clone_id, reversed);V;

  P;
  fprintf(stdout, "[%d] countq reversed: %d\n", my_clone_id, countq(reversed));
  //fprintf(stdout, "[%d] member 0 of reversed: %d\n", my_clone_id, memberq(reversed, 0));
  fprintf(stdout, "[%d] member 1 of reversed: %d\n", my_clone_id, memberq(reversed, 1));
  fprintf(stdout, "[%d] member 2 of reversed: %d\n", my_clone_id, memberq(reversed, 2));
  fprintf(stdout, "[%d] member 11 of reversed: %d\n", my_clone_id, memberq(reversed, 11));
  fprintf(stdout, "[%d] member 12 of reversed: %d\n", my_clone_id, memberq(reversed, 12));
  //fprintf(stdout, "[%d] member 13 of reversed: %d\n", my_clone_id, memberq(reversed, 13));
  V;

  // list deletion is 1-liner, no need to create a delq()
  while (reversed != NULL) (void)popfirstq(reversed);
  while (joined != NULL) (void)popfirstq(joined);
  while (copied != NULL) (void)popfirstq(copied);

  P;showq(my_clone_id, list1);V;
  P;showq(my_clone_id, list2);V;
  P;showq(my_clone_id, joined);V;
  P;showq(my_clone_id, copied);V;
  P;showq(my_clone_id, reversed);V;

  return NULL;
}


#define CLONES 300

/* Unit test of double-ended queue list handling */
int main(int argc, char **argv)
{
  pthread_t pid[CLONES];
  int clone_id[CLONES];
  int my_clone_id = 0; // for consistency with the clone threads, 1..300
  int next_thread;

  pthread_mutex_init(&iolock, NULL);
  pthread_mutex_init(&listlock, NULL);

    next_thread = 0;
    while (next_thread < CLONES) {
      clone_id[next_thread] = next_thread+1;
      if (pthread_create(&pid[next_thread], NULL, list_test, &clone_id[next_thread] /* generic state arg */)) {
        fprintf(stderr, "Failed to spawn thread %d\n", next_thread);
        break;
      }
      next_thread += 1;
    }
    while (next_thread --> 0) pthread_join(pid[next_thread], NULL); // wait for all processes to terminate

  pthread_mutex_destroy(&listlock);
  pthread_mutex_destroy(&iolock);

  exit(EXIT_SUCCESS);
  return EXIT_FAILURE;
}
